/*
 * Copyright (c) 2003-2005 The University of Wroclaw.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *    1. Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *    2. Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *    3. The name of the University may not be used to endorse or promote
 *       products derived from this software without specific prior
 *       written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE UNIVERSITY ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN
 * NO EVENT SHALL THE UNIVERSITY BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

using Nemerle.Compiler;
using Nemerle.Compiler.Typedtree;

namespace Nemerle.Compiler.Typedtree
{
  public variant TypeDeclaration
  {
    | Class
    | Alias         { ty : MType; }
    | Interface
    | Variant       { members : list [TypeInfo]; }
    | VariantOption
    | Enum
  }

  [Record]
  public class Fun_parm : Located
  {
    public name : string;
    public color : int;
    public mutable ty : TyVar;
    public mutable decl : LocalValue;
    public modifiers : Modifiers;
    public kind : ParmKind;

    public SystemType : System.Type
    {
      get {
        ty.Fix ().GetSystemType ();
      }
    }
  }

  public enum FunctionUsage 
  {
    | NotUsed
    | UsedJustOnce
    | Used
    | UsedAsFirstClass
  }
  
  public class Fun_header : Located, Nemerle.IComparable [Fun_header]
  {
    public mutable name : string;
    public id : int;
    public mutable ret_type : TyVar;
    public mutable parms : list [Fun_parm];
    public mutable typarms : list [StaticTyVar];
    public tenv : TyVarEnv;

    public mutable usage : FunctionUsage;
    public mutable body : FunBody;

    /// local values defined in this function, which are closurised (used by children) 
    public mutable closure_vars : list [LocalValue] = [];
    /// local value, which is declaring this function (valid for local funs)
    public mutable decl : LocalValue;
    public mutable all_locals : list [LocalValue] = [];
    internal mutable children_funs : list [Fun_header] = [];

    internal mutable cg_method : CG_method;

    public CompareTo (other : Fun_header) : int
    {
      id - other.id
    }

    public this (loc : Location, 
                 name : string, 
                 ret_type : TyVar, 
                 parms : list [Fun_parm], 
                 typarms : list [StaticTyVar], 
                 tenv : TyVarEnv)
    {
      base (loc);
      this.name = name;
      this.ret_type = ret_type;
      this.parms = parms;
      this.typarms = typarms;
      this.tenv = tenv;

      id = Util.next_id ();
      usage = FunctionUsage.NotUsed;
    }
    
    /**
     *  Retrieves functions in which current function is nested. Empty list if
     *  this is a plain method of some class.
     */
    public GetParents () : list [Fun_header]
    {
      if (decl != null)
        match (decl.ValKind) {
          | LocalValue.Kind.Function (_, pars) => pars
          | _ => Util.ice ("invalid value kind for function")
        }
      else []
    }
    
    public override ToString () : string 
    {
      name
    }
  }

  
  [Record]
  public variant Pattern : Located
  {
    | Wildcard
    | As          { pat : Pattern; decl : LocalValue; }
    | HasType     { tycon : TypeInfo; }
    | Tuple       { args : list [Pattern]; }
    | Record      { args : list [IField * Pattern]; }
    | Application { name : TypeInfo; arg : Pattern; }
    | Enum        { fld : IField; val : Nemerle.Compiler.Literal; }
    | Literal     { lit : Nemerle.Compiler.Literal; }
    | Error

    public mutable ty : TyVar;
    
    public this () { }
  }

  [Record]
  public class Match_case
  {
    public mutable patterns : list [Pattern * TExpr];
    public mutable body : TExpr;
  }

  [Record]
  public variant TExpr : Located
  { 
    | LocalRef              { decl : LocalValue; }
    | StaticRef             { mem : IMember; }
    | ImplicitValueTypeCtor { tc : TypeInfo; }
    | FieldMember           { obj : TExpr; fld : IField; }
    | ConstantObjectRef     { mem : IField; }
    | PropertyMember        { obj : TExpr; prop : IProperty; }
    | StaticPropertyRef     { prop : IProperty; }
    | EventMember           { obj : TExpr; ev : IEvent; }
    | StaticEventRef        { ev : IEvent; }
    | MethodRef             { obj : TExpr; meth : IMethod; notvirtual : bool; }
    | Call                  { mutable func : TExpr; mutable parms : list [Parm]; }
    | TailCall              { func : TExpr; parms : list [Parm]; }
    | SelfTailCall          { parms : list [Parm]; }
    | Assign                { target : TExpr; source : TExpr; }
    | DefValIn              { name : LocalValue; val : TExpr; mutable body : TExpr; }
    | DefFunctionsIn        { funs : list [Fun_header]; mutable body : TExpr; }
    | Match                 { expr : TExpr; cases : list [Match_case]; }
    | Throw                 { exn : TExpr; }
    | TryWith               { body : TExpr; exn : LocalValue; handler : TExpr; }
    | TryFinally            { body : TExpr; handler : TExpr; }
    | Literal               { val : Nemerle.Compiler.Literal; }
    | This
    | Base                  { base_ctor : IMethod; }
    | TypeConversion        { expr : TExpr; target_type : MType; is_checked : bool; }  // (expr :> ty)
    | TypeEnforcement       { expr : TExpr; target_type : MType; } // (expr : ty)
    | Sequence              { e1 : TExpr; e2 : TExpr; }
    | Tuple                 { args : list [TExpr]; }
    | Array                 { args : list [TExpr]; dimensions : list [TExpr]; }
    | TypeOf                { target_type : TyVar; }
    | ArrayIndexer          { over_type : TyVar; obj : TExpr; args : list [TExpr]; }
    | OpCode                { name : string; }
    
    | Delayed               { susp : Typer.DelayedTyping; }
    | Error

    public mutable ty : TyVar;

    public Type : TyVar
    {
      get { assert (ty != null); ty }
    }

    public this () { }
  }

  public enum ParmKind {
    | Normal
    | Ref
    | Out
  }

  [Record]
  public class Parm
  {
    public kind : ParmKind;
    public mutable expr : TExpr;
    public name : string;  // "" if no name given

    public this (expr : TExpr)
    {
      kind = ParmKind.Normal;
      this.expr = expr;
      name = "";
    }
  }
} // Nemerle.Compiler.Typedtree
